%%  FDD/EFDD on a 4 DOF shear frame

% This script generates the dynamic response of a 4 DOF shear frame induced by
% zero mean, Gaussian white noise. The modal paramters are estimated using
% FDD and EFDD.
%
% No external functions are needed to run this script.
%--------------------------------------------------------------------------
%   Structure parameters
K = [1 -1 0 0; -1 2 -1 0; 0 -1 2 -1; 0 0 -1 2];
M = diag([1 1 1 1]);
% The C matrix is calculated by Rayleigh proportional damping
N_dofs = length(K);
%--------------------------------------------------------------------------
% Sample rate and duration of simulation
fs = 10;
f_nyquist = fs/2;
dt = 1/fs; 
n = 2*60*60;
t = 0:dt:n-1/fs;
N = length(t);
%--------------------------------------------------------------------------
% Eigenvalueproblem
[mode_shapes_analytical,lamda] = eig(K,M);
Wn = sqrt(diag(lamda));
fn = Wn./(2*pi);
Tn = (2*pi)./Wn;
%--------------------------------------------------------------------------
% Rayleigh proportional damping
Ksi = [0.01; 0.01];
Rayleigh_matrix = 0.5.*[1/min(Wn) min(Wn); 1/max(Wn) max(Wn)];
an = Rayleigh_matrix^1*Ksi;
% Damping matrx - 
alpha = 0; % Factor for proportional and non-proportional damping
Cp = [0 0 0 1; 0 0 1 0; 0 1 0 0; 0 1 0 0];
C = (an(1).*M+an(2).*K) + alpha.*Cp;
X = C*M^-1*K; % The system is proportionally damped if X is symmetric.
ksi_1 = an(1)/(2*Wn(1))+an(2)*Wn(1)/2;
ksi_2 = an(1)/(2*Wn(2))+an(2)*Wn(2)/2;
ksi_3 = an(1)/(2*Wn(3))+an(2)*Wn(3)/2;
ksi_4 = an(1)/(2*Wn(4))+an(2)*Wn(4)/2;
xi_Rayleigh = [ksi_1 ksi_2 ksi_3 ksi_4];
%--------------------------------------------------------------------------
% White Gaussian noise with randn and defined amplitude
P = 1*[((sqrt(1)*randn(N,1)))';((sqrt(1)*randn(N,1)))';((sqrt(1)*randn(N,1)))';((sqrt(1)*randn(N,1)))'];
%--------------------------------------------------------------------------
%############################ NEWMARKS'S METHOD ###########################
%--------------------------------------------------------------------------
% Gamma=1/2 and Beta=1/4 --> Constant average acceleration method
% Gamma=1/2 and Beta=1/6 --> Linear acceleration method
Gamma=1/2; Beta=1/4;
%Initial displacement, velocity and acceleration.
u0 = zeros(N_dofs,1); 
v0 = zeros(N_dofs,1);
a0 = (M^-1)*(P(:,1)-C*v0-K*u0);

Keff = K+(Gamma/(Beta*dt)).*C+(1/(Beta*(dt)^2)).*M;
ac = (1/(Beta*dt)).*M+(Gamma/Beta).*C; bc = (1/(2*Beta)).*M+dt*((Gamma/(2*Beta))-1).*C;

u = zeros(N_dofs,length(t)); u(:,1) = u0;
v = zeros(N_dofs,length(t)); v(:,1) = v0;
a = zeros(N_dofs,length(t)); a(:,1) = a0;
%--------------------------------------------------------------------------
%   Iterations
for i = 1:(length(t)-1)
    delta_P(:,i+1) = (P(:,i+1)-P(:,i))+ac*v(:,i)+bc*a(:,i);
    delta_u(:,i) = (Keff^-1)*delta_P(:,i+1);
    delta_v(:,i) = (Gamma/(Beta*dt))*delta_u(:,i)-(Gamma/Beta)*v(:,i)+dt*(1-(Gamma/(2*Beta)))*a(:,i);
    delta_a(:,i) = (1/(Beta*(dt)^2))*delta_u(:,i)-(1/(Beta*dt))*v(:,i)-(1/(2*Beta))*a(:,i);
    
    u(:,i+1) = u(:,i)+delta_u(:,i);
    v(:,i+1) = v(:,i)+delta_v(:,i);
    a(:,i+1) = a(:,i)+delta_a(:,i);
end   

% Add noise to the response with mu factor
mu = 0;
u_alt = u + mu * [((sqrt(1)*randn(N,1)))';((sqrt(1)*randn(N,1)))';((sqrt(1)*randn(N,1)))';((sqrt(1)*randn(N,1)))']; 
%--------------------------------------------------------------------------
% ########################## FDD/EFDD METHOD ##############################
%--------------------------------------------------------------------------
%% Calculation of the PSD function
window = N/10; 
noverlap = 0.6*window;
nfft = N;
N_freq = 0.5*nfft+1;

G = zeros(N_dofs,N_dofs,N_freq);
for i = 1:N_dofs;
    for k = 1:N_dofs;
            [G(i,k,:),f] = cpsd(u_alt(i,:),u_alt(k,:),window,noverlap,nfft,fs);
    end
end
% Singular value decompoition
U = zeros(N_dofs,N_dofs,N_freq);
S = zeros(N_dofs,N_dofs,N_freq);
V = zeros(N_dofs,N_dofs,N_freq);
for k = 1:N_freq;
        [U(:,:,k) , S(:,:,k) , V(:,:,k)] = svd(G(:,:,k)); 
end
%% Determine the PSD around the peaks
peak1 = 403; 
peak2 = 1144; 
peak3 = 1748; 
peak4 = 2140; 
mode_estimate_1 = U(:,1,peak1);
mode_estimate_2 = U(:,1,peak2);
mode_estimate_3 = U(:,1,peak3);
mode_estimate_4 = U(:,1,peak4);
MODE_ESTIMATE = [mode_estimate_1 mode_estimate_2 mode_estimate_3 mode_estimate_4];
sigma = zeros(N_dofs,N_freq); % Zero vector for the singular values
OMEGA = 0.8;
for i = 1:N_dofs; 
    if i == 1;
        j = peak1;
    elseif i == 2;
        j = peak2;
    elseif i == 3;
        j = peak3;
    elseif i == 4;
        j = peak4;
    end 
        for n = 1:N_dofs; 
            for k = j:1:N_freq; 
                    A = abs(U(:,n,k)'*MODE_ESTIMATE(:,i))^2; 
                    B = (U(:,n,k)'*U(:,n,k))*(MODE_ESTIMATE(:,i)'*MODE_ESTIMATE(:,i));
                    MAC = A/B;
                        if MAC > OMEGA;
                        sigma(i,k) = sigma(i,k) + S(n,n,k);
                        elseif MAC == OMEGA;
                        sigma(i,k) = sigma(i,k) + S(n,n,k);
                        elseif MAC < OMEGA;
                        break
                        end
            end
            for k = j-1:-1:1; 
                    A = abs(U(:,n,k)'*MODE_ESTIMATE(:,i))^2; 
                    B = (U(:,n,k)'*U(:,n,k))*(MODE_ESTIMATE(:,i)'*MODE_ESTIMATE(:,i));
                    MAC = A/B;
                        if MAC > OMEGA;
                        sigma(i,k) = sigma(i,k) + S(n,n,k);
                        elseif MAC == OMEGA;
                        sigma(i,k) = sigma(i,k) + S(n,n,k);
                        elseif MAC < OMEGA;
                        break
                        end
            end
        end
end

G_SDOF = sigma;
corr_SDOF = zeros(N_dofs,length(sigma));
for j = 1:N_dofs;
   corr(j,:) = (ifft(G_SDOF(j,:),'symmetric')); 
end

corr_norm = zeros(N_dofs,length(sigma));
for k = 1:N_dofs;
    corr_norm(k,:) = (corr(k,:)-mean(corr(k,:)))/corr(k,1);
end
% Estimating the natural frequency of the autocorrelation functions
[peak_SDOF1 , loc_SDOF1] = findpeaks(abs(corr_norm(1,:)));
[peak_SDOF2 , loc_SDOF2] = findpeaks(abs(corr_norm(2,:)));
[peak_SDOF3 , loc_SDOF3] = findpeaks(abs(corr_norm(3,:)));
[peak_SDOF4 , loc_SDOF4] = findpeaks(abs(corr_norm(4,:)));
% Converting bins to time after FFT
time = 0:(2/fs):N/fs;
length(time);
%% Damping estimation
delta1 = zeros(1,length(peak_SDOF1));
for k = 1:length(peak_SDOF1)
    delta1(k) = (2/k) * log(corr_norm(1,1)/abs(peak_SDOF1(k)));
end

delta2 = zeros(1,length(peak_SDOF2));
for k = 1:length(peak_SDOF2)
    delta2(k) = (2/k) * log(corr_norm(2,1)/abs(peak_SDOF2(k)));
end

delta3 = zeros(1,length(peak_SDOF3));
for k = 1:length(peak_SDOF3)
    delta3(k) = (2/k) * log(corr_norm(3,1)/abs(peak_SDOF3(k)));
end

delta4 = zeros(1,length(peak_SDOF4));
for k = 1:length(peak_SDOF4)
    delta4(k) = (2/k) * log(corr_norm(4,1)/abs(peak_SDOF4(k)));
end

j = 6;
Y_delta1 = delta1.*loc_SDOF1 + 2*log(abs(peak_SDOF1));
P1 = polyfit(loc_SDOF1(1:j),Y_delta1(1:j),1);
Y1 = P1(1)*loc_SDOF1 + P1(2);
xi1 = P1(1)/sqrt(P1(1)^2+(2*pi)^2);

k = 17;
Y_delta2 = delta2.*loc_SDOF2 + 2*log(abs(peak_SDOF2));
P2 = polyfit(loc_SDOF2(1:k),Y_delta2(1:k),1);
Y2 = P2(1)*loc_SDOF2 + P2(2);
xi2 = P2(1)/sqrt(P2(1)^2+(2*pi)^2);

m = 29;
Y_delta3 = delta3.*loc_SDOF3 + 2*log(abs(peak_SDOF3));
P3 = polyfit(loc_SDOF3(1:m),Y_delta3(1:m),1);
Y3 = P3(1)*loc_SDOF3 + P3(2);
xi3 = P3(1)/sqrt(P3(1)^2+(2*pi)^2);

n = 30;
Y_delta4 = delta4.*loc_SDOF4 + 2*log(abs(peak_SDOF4));
P4 = polyfit(loc_SDOF4(1:n),Y_delta4(1:n),1);
Y4 = P4(1)*loc_SDOF4 + P4(2);
xi4 = P4(1)/sqrt(P4(1)^2+(2*pi)^2);

xi_estimate = [xi1 xi2 xi3 xi4];
%% Frequency estimation
%Number of upp crossings in the correlation functions
count = zeros(N_dofs,1);
count_loc = zeros(N_dofs,length(corr_norm(N_dofs,:)));
for n = 1:N_dofs;
    for j = 1:length(corr_norm(1,:))-1;
     if corr_norm(n,j) < 0 && corr_norm(n,j+1) > 0;
            count(n,1) = count(n,1) +1;
            count_loc(n,j) = count_loc(n,j) + (count(n,1) + 1);
     elseif corr_norm(n,j) < 0 && corr_norm(n,j+1) == 0
            continue
        elseif corr_norm(n,j) < 0 && corr_norm(n,j+1) < 0
            continue
        elseif corr_norm(n,j) == 0 && corr_norm(n,j+1) > 0
            continue
        elseif corr_norm(n,j) == 0 && corr_norm(n,j+1) > 0
            continue
        elseif corr_norm(n,j) == 0 && corr_norm(n,j+1) == 0
            continue
        elseif corr_norm(n,j) > 0 && corr_norm(n,j+1) > 0
            continue
        elseif corr_norm(n,j) > 0 && corr_norm(n,j+1) == 0
            continue
        elseif corr_norm(n,j) > 0 && corr_norm(n,j+1) < 0
           continue
     end
    end
end

j = 10;
[peaks_zc1, loc_zc1] = findpeaks(count_loc(1,:));
P_zc1 = polyfit(loc_zc1(1:j),peaks_zc1(1:j),1);
Y_zc1 = P_zc1(1)*loc_zc1 + P_zc1(2);
fd1_estimate = P_zc1(1)*0.5*fs;
fn1_estimate = fd1_estimate/sqrt(1-xi1^2);

k = 30;
[peaks_zc2, loc_zc2] = findpeaks(count_loc(2,:));
P_zc2 = polyfit(loc_zc2(1:k),peaks_zc2(1:k),1);
Y_zc2 = P_zc2(1)*loc_zc2 + P_zc2(2);
fd2_estimate = P_zc2(1)*0.5*fs;
fn2_estimate = fd2_estimate/sqrt(1-xi2^2);

m = 24;
[peaks_zc3, loc_zc3] = findpeaks(count_loc(3,:));
P_zc3 = polyfit(loc_zc3(1:m),peaks_zc3(1:m),1);
Y_zc3 = P_zc3(1)*loc_zc3 + P_zc3(2);
fd3_estimate = P_zc3(1)*0.5*fs;
fn3_estimate = fd3_estimate/sqrt(1-xi3^2);

n = 28;
[peaks_zc4, loc_zc4] = findpeaks(count_loc(4,:));
P_zc4 = polyfit(loc_zc4(1:n),peaks_zc4(1:n),1);
Y_zc4 = P_zc4(1)*loc_zc4 + P_zc4(2);
fd4_estimate = P_zc4(1)*0.5*fs;
fn4_estimate = fd4_estimate/sqrt(1-xi4^2);

fd_estimate = [fd1_estimate fd2_estimate fd3_estimate fd4_estimate];
fn_estimate = [fn1_estimate fn2_estimate fn3_estimate fn4_estimate];
%--------------------------------------------------------------------------
%% ########################### PLOTS ######################################
%--------------------------------------------------------------------------
%%  Singular values
figure(2);
semilogy(squeeze(S(1,1,:))); grid; hold on;
semilogy(squeeze(S(2,2,:)));
semilogy(squeeze(S(3,3,:)));
semilogy(squeeze(S(4,4,:))); hold on;